import argparse


def custom_formatter(prog):
    return argparse.HelpFormatter(prog, max_help_position=59, width=125)


def cli():
    parser = argparse.ArgumentParser(
        description     = "declaratively configure PKI secrets in Hashicorp Vault",
        formatter_class = custom_formatter
    )

    parser.add_argument('-d', '--debug', dest='debugging',
        action='store_true', default=False, help='enable debug output')
    parser.add_argument('-V', '--version', action='version', version='Vault-PKI 0.1')

    subparsers = parser.add_subparsers(title='subcommands', dest='subcommand', metavar='')

    init = subparsers.add_parser(
        'init',
        help="Initializes the Hashicorp Vault server",
        formatter_class=custom_formatter
    )

    init.add_argument('-u', '--url', dest='baseurl', type=str, metavar='URL',
        action='store', required=False, help='the URL of the Vault server')
    init.add_argument('--tls-skip-verify', nargs='?', dest='tls_skip_verify',
        const=True, default=None, help="disable verification of the Vault server's SSL certificate")

    apply = subparsers.add_parser(
        'apply',
        help="Creates PKI secrets from a YAML file",
        formatter_class=custom_formatter
    )

    apply.add_argument('-u', '--url', dest='baseurl', type=str, metavar='URL',
        action='store', required=False, help='the URL of the Vault server')
    apply.add_argument('-f', '--file', dest='file', type=str,
        action='store', required=True, help='the path to the configuration manifest(s)')
    apply.add_argument('--tls-skip-verify', nargs='?', dest='tls_skip_verify',
        const=True, default=None, help="disable verification of the Vault server's SSL certificate")

    return parser



from . import utils
from urllib.parse import urljoin
import requests


class VaultClient:
    def __init__(self, baseurl=None, token=None, verify_ssl=True, debugging=False):
        self.baseurl     = baseurl
        self.token       = token
        self.verify_ssl  = verify_ssl
        self.debugging   = debugging
        self.timeout     = 80
        self.master_keys = []

    @property
    def headers(self):
        return {'X-VAULT-TOKEN': self.token}

    def request(self, method, url, headers=None, json=None):
        try:
            response = requests.request(
                method=method,
                url=url,
                headers=headers,
                json=json,
                timeout=self.timeout,
                verify=self.verify_ssl
            )
        except requests.exceptions.RequestException as err:
            utils.exit_with_message(f"Failed to contact the Vault server: {err}")
        else:
            if self.debugging:
                msg = f"Request method: {method}, Request URL: {url}, Response status code: {response.status_code}, Response body: {response.text}"
                utils.output_message(msg)

            # See: https://www.vaultproject.io/api/index.html#http-status-codes
            if response.status_code == 403:
                utils.exit_with_message("Failed to authenticate to the Vault server: invalid token")
            elif response.status_code == 404:
                utils.exit_with_message("Failed to process request: invalid path")
            return response

    def healthcheck(self):
        """ checks if the Vault server has been initialized and is not sealed """
        URL = urljoin(self.baseurl, "v1/sys/health")

        response = self.request(method='GET', url=URL)
        body     = response.json()

        initialized = body['initialized']
        sealed      = body['sealed']

        if response.status_code == 200:
            utils.output_message("the Vault server has been initialized and is not sealed")
        elif response.status_code == 501:
            utils.output_message("the Vault server has not been initialized")
        elif response.status_code == 503:
            utils.output_message("the Vault server is sealed", err=True)
        return initialized, sealed

    def initialize_server(self, log_file='vault.log', token_file='.vault-token'):
        """" initializes the Vault server and writes master & root key to disk """
        URL = urljoin(self.baseurl, "v1/sys/init")

        params = {
            "secret_shares": 5,
            "secret_threshold": 3
        }

        response = self.request(method='PUT', url=URL, json=params)

        if response.status_code == 200:
            body = response.json()

            self.master_keys = body['keys_base64']
            self.token = body['root_token']

            utils.write_vault_master_keys(self.master_keys, log_file, self.debugging)
            utils.write_vault_root_token(self.token, token_file, self.debugging)

            utils.output_message("Initialized the Vault server")
        else:
            utils.output_message("failed to initialize the Vault server", err=True)

    def unseal_server(self):
        """" unseals the Vault server """
        URL = urljoin(self.baseurl, "v1/sys/unseal")

        for i in self.master_keys:
            response = self.request(method='PUT', url=URL, json={'key': i})

            if response.status_code == 200:
                body = response.json()

                if not body['sealed']:
                    utils.output_message("Unsealed the Vault server")
                    break
            else:
                utils.exit_with_message("failed to unseal the Vault server")

    def mount_kv_engine(self, kvengine):
        """ mounts a KV v1 secrets engine """
        response = self.request(method='POST', url=kvengine.url, headers=self.headers, json=kvengine.spec)

        if response.status_code == 204:
            utils.output_message(f"Mounted KV secrets engine: {kvengine.name}")
        elif response.status_code == 400:
            utils.output_message(f"KV secrets engine '{kvengine.name}' already exists")
        else:
            utils.exit_with_message(f"Failed to mount KV secrets engine: {kvengine.name}")

    def store_ca_private_key(self, ca):
        """ stores the private key for a CA in the specified KV engine """

        response = self.request(method='PUT', url=ca.kv_backend_url, headers=self.headers, json=ca.private_key)

        if response.status_code == 204:
            utils.output_message(f"Stored private key for '{ca.name}' in KV engine: {ca.kv_backend}")
        else:
            utils.exit_with_message(f"Failed to store private key for '{ca.name}' in KV engine: {ca.kv_backend}")

    def mount_pki_engine(self, ca):
        """ mounts a PKI secrets engine """
        URL = urljoin(self.baseurl, f"v1/sys/mounts/{ca.name}")

        response = self.request(method='POST', url=URL, headers=self.headers, json=ca.backend)

        if response.status_code == 204:
            utils.output_message(f"Mounted PKI secrets engine: {ca.name}")
        elif response.status_code == 400:
            utils.output_message(f"PKI secrets engine '{ca.name}' already exists")
        else:
            utils.exit_with_message(f"Failed to mount PKI secrets engine: {ca.name}")

    def check_existing_ca(self, ca, quiet=False):
        """ checks if a CA already exists """
        ca_exists = False

        URL = urljoin(self.baseurl, f"/v1/{ca.name}/ca/pem")

        response = self.request(method='GET', url=URL, headers=self.headers)

        if response.status_code == 200:
            ca_exists = True
            if not quiet:
                utils.output_message(f"CA '{ca.name}' already exists")
        return ca_exists

    def create_root_ca(self, ca):
        """ generates a Root CA """
        URL = urljoin(self.baseurl, f"/v1/{ca.name}/root/generate/internal")

        response = self.request(method='POST', url=URL, headers=self.headers, json=ca.spec)

        if response.status_code == 200:
            body = response.json()
            if body['data'] is not None:
                utils.output_message(f"Generated Root CA: {ca.name}")
            else:
                utils.output_message(f"Root CA '{ca.name}' has already been generated")
        elif response.status_code == 204:
            utils.output_message(f"Root CA '{ca.name}' has already been generated")
        else:
            utils.exit_with_message(f"Failed to generate Root CA: {ca.name}")

    def configure_ca_urls(self, ca):
        """ configures URLs for a CA """
        response = self.request(method='POST', url=ca.config_url, headers=self.headers, json=ca.ca_urls)

        if response.status_code == 204:
            utils.output_message(f"Configured URLs for CA: {ca.name}")
        else:
            utils.exit_with_message(f"Failed to configure URLs for CA: {ca.name}")

    def set_crl_configuration(self, ca):
        """ sets the duration for CRL validity """
        response = self.request(method='POST', url=ca.crl_config_url, headers=self.headers, json=ca.crl_config)

        if response.status_code == 204:
            utils.output_message(f"Set CRL configuration for CA: {ca.name}")
        else:
            utils.exit_with_message(f"Failed to set CRL configuration for CA: {ca.name}")

    def create_intermediate_ca(self, ca):
        """ generates an Intermediate CA """
        response = self.request(method='POST', url=ca.url, headers=self.headers, json=ca.spec)

        if response.status_code == 200:
            body = response.json()
            ca.csr = body['data']['csr']

            if ca.catype == 'exported':
                ca.private_key = body['data']['private_key']

            utils.output_message(f"Created intermediate CA: {ca.name}")
        else:
            utils.exit_with_message(f"Failed to generate intermediate CA: {ca.name}")

    def sign_intermediate_ca(self, ca):
        """ signs the certificate for an Intermediate CA with another CA """
        spec = ca.spec.copy()
        spec.update(csr=ca.csr)

        response = self.request(method='POST', url=ca.issuer_sign_url, headers=self.headers, json=spec)

        if response.status_code == 200:
            body        = response.json()
            certificate = body['data']['certificate']

            # there is no ca_chain for Root CAs
            try:
                ca_chain = '\n'.join(body['data'].get('ca_chain'))
            except TypeError:
                ca_chain = body['data']['issuing_ca']

            # include the full chain with the signed certificate
            ca.cert = f"{certificate}\n{ca_chain}"

            utils.output_message(f"Signed intermediate CA '{ca.name}' with issuing CA: {ca.issuer}")
        else:
            utils.exit_with_message(f"Failed to sign intermediate CA '{ca.name}' with issuing CA: {ca.issuer}")

    def set_intermediate_ca(self, ca):
        """ sets the signed certificate for an Intermediate CA """

        params = {'certificate': ca.cert}

        response = self.request(method='POST', url=ca.set_signed_url, headers=self.headers, json=params)

        if response.status_code == 204:
            utils.output_message(f"Set signed certificate for intermediate CA: {ca.name}")
        else:
            utils.exit_with_message(f"Failed to set signed certificate for intermediate CA: {ca.name}")

    def configure_ca_roles(self, ca):
        """ configures roles for an Intermediate CA """
        for role in ca.roles:
            name = role['name']
            config = role['config']

            URL = urljoin(self.baseurl, f"/v1/{ca.name}/roles/{name}")

            response = self.request(method='POST', url=URL, headers=self.headers, json=config)

            if response.status_code == 204:
                utils.output_message(f"Configured role '{name}' for intermediate CA: {ca.name}")
            else:
                utils.exit_with_message(f"Failed to configure role '{name}' for intermediate CA: {ca.name}")

    def configure_ca_policies(self, ca):
        for policy in ca.policies:
            name = policy['name']
            document = {'policy': policy['policy']}

            URL = urljoin(self.baseurl, f"/v1/sys/policies/acl/{name}")

            response = self.request(method='PUT', url=URL, headers=self.headers, json=document)

            if response.status_code == 204:
                utils.output_message(f"Configured policy '{name}' for intermediate CA: {ca.name}")
            else:
                utils.exit_with_message(f"Failed to configure policy '{name}' for intermediate CA: {ca.name}")



from . import schemas
from typing import List, Tuple, Optional
import getpass
import glob
import os
import sys
import yaml


def output_message(msg: str, err: bool = False):
    prefix = "[-]" if err else "[*]"
    message = [prefix, 'pkictl', '-', msg]
    if err:
        message.insert(3, "Error:")
    print(' '.join(message), flush=True)


def exit_with_message(msg: str):
    sys.exit(f"[-] pkictl - Error: {msg}")


def get_from_environment(name: str):
    value = os.getenv(name)
    if name == 'VAULT_ADDR' and value is None:
        return input("Vault URL: ")
    elif name == 'VAULT_TOKEN' and value is None:
        return getpass.getpass('Vault Token: ')
    elif name == 'VAULT_SKIP_VERIFY' and value is None:
        return 'False'
    return value


def get_manifest_files(directory: str) -> List[str]:
    """ returns a list of absolute paths to YAML manifest files within a directory """
    pattern  = '*.y[am]*l'  # match .yaml or .yml
    return glob.glob(f'{directory}/{pattern}')


def read_manifest_file(path: str) -> List[dict]:
    try:
        with open(path, 'r') as f:
            return list(yaml.load_all(f.read()))
    except FileNotFoundError:
        error_message = "manifest file does not exist"
    except yaml.YAMLError:
        error_message = "failed to parse manifest file, invalid YAML"
    except PermissionError:
        error_message = "failed to read manifest file, permission denied"
    except Exception as err:
        error_message = f"failed to read manifest file. Exception: {err}"
    return exit_with_message(error_message)


def get_validated_manifests(documents: List[dict]=[]) -> Tuple[List[dict], List[dict], List[dict]]:
    roots: List[dict]           = []
    intermediates: List[dict]   = []
    kv_backends: List[dict]     = []

    for i in documents:
        schema_type = i.get('kind')

        if schema_type == 'RootCA':
            roots.append(schemas.RootCASchema(i))

        elif schema_type == 'IntermediateCA':
            ca_name     = i['name']
            ca_type     = i['spec'].get('type')
            kv_backend  = i.get('kv_backend', None)

            if ca_type == 'exported' and kv_backend is None:
                exit_with_message(f"kv_backend not defined for exported intermediate CA: {ca_name}")

            intermediates.append(schemas.IntermediateCASchema(i))

        elif schema_type == 'KV':
            kv_backends.append(schemas.KeyValueSchema(i))

        else:
            exit_with_message("Unsupported schema defined in manifest file")
    return roots, intermediates, kv_backends


def write_vault_master_keys(master_keys: List[str]=[], file: str='vault.log', debug: bool=False) -> None:
    try:
        with open(file, 'w') as f:
            for i, v in enumerate(master_keys, 1):
                f.write(f"Unseal Key {i}: {v}\n")
    except Exception:
        exit_with_message(f"Failed to write the Vault master keys to {file}")
    else:
        os.chmod(file, 0o600)
        if debug:
            output_message(f"Successfully wrote the Vault master keys to {file}")


def write_vault_root_token(root_token: str=None, file: str='.vault-token', debug: bool=False) -> None:
    try:
        with open(file, 'w') as f:
            f.write(f"{root_token}\n")
    except Exception:
        exit_with_message(f"Failed to write the Vault root token to {file}")
    else:
        os.chmod(file, 0o600)
        if debug:
            output_message(f"Successfully wrote the Vault root token to {file}")


def get_index(array: List[dict], key: str, value: str) -> Optional[int]:
    for i, d in enumerate(array):
        if d[key] == value:
            return i
    return None


def sort_intermediate_certificate_authorities(intermediates: List[dict]):
    sorted_intermediates: List[dict] = []
    for ca in intermediates:
        issuer = ca.get('issuer')

        if issuer:
            index  = get_index(intermediates, key='name', value=issuer)

            if index is None:
                sorted_intermediates.insert(0, ca)
            else:
                sorted_intermediates.insert(index, ca)
    return sorted_intermediates



from setuptools import setup

version = "0.1.2"

requirements = [
    'PyYAML',
    'requests',
    'voluptuous'
]

with open("README.md", "r") as f:
    readme = f.read()

setup(name='pkictl',
    version=version,
    description='A command-line utility for declaratively provisioning PKI in Hashicorp Vault',
    long_description=readme,
    long_description_content_type="text/markdown",
    url='http://github.com/bincyber/pkictl',
    author='@bincyber',
    license='MPL',
    keywords="vault pki kubernetes public key infrastructure security",
    packages=['pkictl'],
    python_requires='>=3.6',
    platforms=['any'],
    install_requires=requirements,
    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Operating System :: MacOS',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Security',
        'Topic :: Security :: Cryptography',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities'
    ],
    entry_points = {
        'console_scripts': ['pkictl=pkictl.pkictl:main']
    },
    test_suite='nose2.collector.collector',
    tests_require=['nose2']
)
